hitmds <- function(MD, Y, n_dim, rate, n_cycles) {

  if(!is.matrix(MD))
    stop('MD no matrix!')

  if(!isSymmetric(MD))
    stop('Distance matrix MD not symmetric!')

  n_data <- nrow(MD)
  n_datm <- n_data / (n_data-1)
  n_iter <- n_cycles * n_data
  n_itr1 <- n_iter 
  n_itr2 <- n_iter / 2

  if(n_dim == 1 && length(Y) > 0 && !is.matrix(Y))
    Y <- as.matrix(Y)

  if(length(Y) > 0 && (nrow(Y) != n_data || ncol(Y) != n_dim))
    stop('Y matrix has wrong size!')

  n_elem <- n_data * (n_data-1) * .5

  if(length(Y) == 0)
    Y <- 2 * (matrix(runif(n_data*n_dim), n_data, n_dim) - .5)

  MT <- as.matrix(dist(Y, diag=T, upper=T))

  Y <- t(Y)

  cnt <- 1:n_data
  eps <- 1e-16
  
  mn_T <- n_datm * mean(MT)
  tmp <- MT - mn_T
  MD <- MD - n_datm * mean(MD)
  mi_T <- .5 * (sum(tmp *  MD) - sum(diag(tmp) * diag(MD)))
  mo_T <- .5 * (sum(tmp * tmp) - sum(diag(tmp) * diag(tmp)))
  mo_D <- .5 * (sum( MD *  MD) - sum(diag( MD) * diag(MD)))

  i_rep <- 0

  while(n_iter > 0) {

    i_rep <- i_rep - 1

    if(i_rep <= 0) {
      cat(paste(n_cycles,sqrt(max(mi_T*mi_T,eps)/(mo_D*mo_T)),sep=" "),sep="\n")
      n_cycles <- n_cycles - 1
      i_rep <- n_data
      shuffl <- sample(cnt)
    }

    i <- shuffl[i_rep]
    pnt <- Y[,i]
    sq <- cnt[-i]

    tmp <- (MT[sq,i]-mn_T) * mi_T - MD[sq,i] * mo_T
    tmp <- tmp / pmax(MT[sq,i], eps) * t(pnt - Y[,sq]) 
    pnt_del <- apply(tmp, 2, sum)
    rt <- ifelse(n_iter<n_itr2, rate * (1 + 2 * (n_iter-n_itr2)/n_itr1), rate)
    cat(rt)
    pnt_del <- pnt - rt * pnt_del / pmax(abs(pnt_del),eps)

    tmp <- sqrt(apply((pnt_del-Y[,sq])^2, 2, sum))
    ddif_ <- tmp - MT[sq, i]
    dif <- sum(ddif_)

    d_mn <- dif / n_elem
    d_of <- -d_mn - mn_T

    d_mi <- sum(MD[sq,i] * ddif_)
    d_mo <- sum(ddif_ * (ddif_ + 2 * (MT[sq,i] + d_of)))

    mn_T <- mn_T + d_mn
    mi_T <- mi_T + d_mi
    mo_T <- mo_T + d_mo + n_elem * d_mn * d_mn

    MT[sq,i] <- MT[sq,i] + ddif_
    MT[i,sq] <- t(MT[sq,i])

    Y[,i] <- pnt_del

    n_iter <- n_iter - 1

  }
 
  cat(paste(0,sqrt(max(mi_T*mi_T,eps) / (mo_D * mo_T)),sep=" "),sep="\n")
  
  t(Y)
}

stdscatter <- function(y) {
  y <- (t(y) - apply(y, 2, mean))
  y <- crossprod(svd(y, nv = 0)$u, y)
  t(y * sign(apply(sign(y) * y * y, 1, sum))) / max(apply(y, 1, sd))
}

MD <- as.matrix(dist(dat,diag=T,upper=T))
d_goal <- 2
l_rate <- .1
n_cyc  <- 50
d_inp  <- ncol(dat)
y <- dat %*% (2 * (matrix(runif(d_inp * d_goal), d_inp, d_goal) - .5))
y <- hitmds(MD, y, d_goal, l_rate, n_cyc)
stdscatter(y)